/*
** Module   : DatetimeTzUserType.java
** Abstract : Custom user type to wrap P2J datetime for Hibernate access
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20130122 Created initial version. Maps P2J datetimetz to SQL
**                  timestamp and smallint columns.  2 columns are needed
**                  because the Progress TZ is persisted as an integer offset
**                  which cannot be stored in a SQL timestamp alone.
** 002 OM  20130524 Rewrote entire class to extend CompositeUserType in order to accommodate the 
**                  compound type.
*/

package com.goldencode.p2j.persist.type;

import java.io.Serializable;
import java.sql.*;

import org.apache.commons.logging.*;
import org.hibernate.HibernateException;
import org.hibernate.engine.spi.SessionImplementor;
import org.hibernate.internal.util.StringHelper;
import org.hibernate.type.*;
import org.hibernate.usertype.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of a custom, Hibernate user type for the {@link
 * com.goldencode.p2j.util.datetimetz P2J datetimetz}. This permits Hibernate to
 * convert this data type to and from the SQL TIMESTAMP and SMALLINT aggregate
 * data types. The unknown value is mapped to SQL NULL in the TIMESTAMP column.
 */
public final class DatetimeTzUserType
implements CompositeUserType
{
   /** Two element array containing the HIBERNATE type being mapped. */
   private static final Type[] HIBERNATE_TYPES = { TimestampType.INSTANCE, IntegerType.INSTANCE };
   
   /** Logger */
   private static final Log LOG = LogFactory.getLog(StringHelper.qualifier(Type.class.getName()));
   
   /** Is trace level logging enabled? */
   private static final boolean LOG_TRACE = LOG.isTraceEnabled();
   
   /**
    * Get the "property names" that may be used in a query.
    *
    * @return  An array of Strings that contains the names of the fields occupied by this object. 
    */
   @Override
   public String[] getPropertyNames()
   {
      return new String[] { "timestamp", "offset" };
   }

   /**
    * Get the corresponding "property types".
    *
    * @return an array of Hibernate types
    */
   @Override
   public Type[] getPropertyTypes()
   {
      return HIBERNATE_TYPES;
   }

   /**
    * Get the value of a property.
    *
    * @param   component
    *          An instance of class mapped by this "type".
    * @param   property
    *          Index of property to be returned.
    * 
    * @return  The property value
    * 
    * @throws HibernateException
    */
   @Override
   public Object getPropertyValue(Object component, int property)
   throws HibernateException
   {
      if (component == null || !(component instanceof datetimetz))
      {
         return null;
      }
      
      datetimetz dtz = (datetimetz) component;
      if (property == 0)
      {
         return new Timestamp(dtz.dateValue().getTime());
      }
      else if (property == 1)
      {
         return dtz.getOffset();
      }

      return null;
   }

   /**
    * Set the value of a property.
    *
    * @param   component
    *          An instance of datetimetz class mapped by this "type".
    * @param   property
    *          Index of property to be set.
    * @param   value
    *          The value to set.
    * 
    * @throws HibernateException
    */
   @Override
   public void setPropertyValue(Object component, int property, Object value)
   throws HibernateException
   {
      if (value == null || component == null || !(component instanceof datetimetz))
      {
         return;
      }
      
      datetimetz dtz = (datetimetz) component;
      if (property == 0)
      {
         dtz.setTimestamp((Date) value); // a Timestamp in fact
      }
      else if (property == 1)
      {
         dtz.setTimeZoneOffset((Integer) value);
      }
   }

   /**
    * Tests for equality between two data wrapper instances. The objects are considered the same
    * if they meet any of the following conditions:
    * <ul>
    *    <li>they represent the same object instance in memory; OR
    *    <li>they are both <code>null</code>; OR
    *    <li><code>x.equals(y)</code> evaluates to <code>true</code>.
    * </ul>
    * Otherwise, they are considered different.
    * <p>
    * This method is used by Hibernate for dirty checking.
    *
    * @param   x
    *          A data wrapper type instance.
    * @param   y
    *          A data wrapper type instance.
    */
   @Override
   public boolean equals(Object x, Object y)
   throws HibernateException
   {
      if (x == y)
      {
         return true;
      }

      if (x == null || y == null)
      {
         return false;
      }

      return x.equals(y);
   }

   /**
    * Get the hash code of the given object in a manner that is consistent with the equality check
    * performed by {@link #equals}. This implementation simply returns the result of invoking
    * <code>x.hashCode()</code>.
    *
    * @param   x
    *          A data wrapper type instance.
    *
    * @return  The given object's hash code.
    */
   @Override
   public int hashCode(Object x)
   throws HibernateException
   {
      return x.hashCode();
   }

   /**
    * Retrieve an instance of the <code>datetimetz</code> class from a JDBC resultset. 
    *
    * @param   rs 
    *          A JDBC result set.
    * @param   names
    *          The column names.
    * @param   sess
    * @param   owner
    *          The containing entity
    *          
    * @return  The instantiated <code>datetimetz</code> object.
    *  
    * @throws  HibernateException
    * @throws  SQLException
    *          if a database access error occurs.
    */
   @Override
   public Object nullSafeGet(ResultSet rs, String[] names, SessionImplementor sess, Object owner)
   throws HibernateException, SQLException
   {
      Date timestamp = rs.getDate(names[0]);
      int offset     = rs.getShort(names[1]);
      
      datetimetz ret = new datetimetz();
      ret.setTimeZoneOffset(offset);
      ret.setTimestamp(timestamp);
      return ret;
   }
   
   /**
    * Write an instance of the <code>datetimetz</code> class to a prepared statement. 
    * A multi-column type should be written to parameters starting from <tt>index</tt>.
    *
    * @param   st
    *          A JDBC prepared statement.
    * @param   value
    *          The object to be written.
    * @param   index
    *          The statement parameter index.
    * @param   sess
    * 
    * @throws HibernateException
    * @throws SQLException
    */
   @Override
   public void nullSafeSet(PreparedStatement st, Object value, int index, SessionImplementor sess)
   throws HibernateException, SQLException
   {
      if (value == null)
      {
         st.setObject(index, null);
         st.setObject(index + 1, null);
      }
      
      datetimetz dtz = (datetimetz)value;
      if (dtz.isUnknown())
      {
         st.setObject(index, null);
         st.setObject(index + 1, null);
      }
      else
      {
         st.setTimestamp(index, new Timestamp(dtz.dateValue().getTime()));
         st.setInt(index + 1, dtz.getOffset());
      }
   }

   /**
    * Returns <code>true</code> to indicate that the P2J data wrapper object is mutable.
    *
    * @return  <code>true</code>.
    */
   @Override
   public boolean isMutable()
   {
      return true;
   }

   /**
    * Transform the object into its cacheable representation. This method returns a deep copy 
    * of the passed object. 
    *
    * @param   value
    *          The object to be cached.
    * @param   session
    * 
    * @return  A cacheable representation of the object.
    * 
    * @throws HibernateException
    */
   @Override
   public Serializable disassemble(Object value, SessionImplementor session)
   throws HibernateException
   {
      return (Serializable)deepCopy(value);
   }

   /**
    * Reconstruct an object from the cacheable representation. 
    *
    * @param   cached
    *          The object to be cached.
    * @param   session
    * @param   owner
    *          The owner of the cached object.
    * 
    * @return  a reconstructed object from the cacheable representation.
    * 
    * @throws HibernateException
    */
   @Override
   public Object assemble(Serializable cached, SessionImplementor session, Object owner)
   throws HibernateException
   {
      return deepCopy(cached);
   }

   /**
    * During merge, replace the existing (target) value in the entity we are merging to
    * with a new (original) value from the detached entity we are merging. For immutable
    * objects, or null values, it is safe to simply return the first parameter. For
    * mutable objects, it is safe to return a copy of the first parameter. However, since
    * composite user types often define component values, it might make sense to recursively 
    * replace component values in the target object.
    * 
    * @param original
    * @param target
    * @param session
    * @param owner
    * 
    * @return
    * 
    * @throws HibernateException
    */
   @Override
   public Object replace(Object original, Object target, SessionImplementor session, Object owner)
   throws HibernateException
   {
      return deepCopy(original);
   }
   
   /**
    * Gets the Java data type class mapped by this custom type.
    *
    * @return  The {@link com.goldencode.p2j.util.datetimetz P2J datetimetz} class.
    */
   @Override
   public Class returnedClass()
   {
      return datetimetz.class;
   }
   
   /**
    * Create and return a deep copy of the given <code>datetimetz</code> instance.
    *
    * @param   value
    *          The <code>datetimetz</code> instance to be copied.
    *
    * @return  A new instance of <code>datetimetz</code> which contains the
    *          same internal value as <code>value</code>.
    */
   @Override
   public Object deepCopy(Object value)
   {
      return new datetimetz((datetimetz) value);
   }
}
